#
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:

require 'strscan'

require 'sfllib'

module SFL
	KEYWORDS = {}
	%w{alt any bidirect bus bus_v call circuit declare else finish first_state generate goto if input instr_arg instrin instrout instrself instruct mem module output par reg reg_wr reg_ws relay return segment segment_name sel sel_v stage stage_name state state_name task}.each do|w|
		KEYWORDS[w] = w.upcase.to_sym
	end

	RESERVED = {}
	%w{p_reset m_clock s_clock b_clock VDD VSS scan_in scan_out scan_enb scan_clock}.each do|w|
		RESERVED[w] = true
	end

	DBOPS = {
		'/&' => :OP_FOLDAND,
		'/@' => :OP_FOLDEOR,
		'/|' => :OP_FOLDOR,
		':=' => :OP_ASSIGNREG,
		'<<' => :OP_SHIFTLEFT,
		'==' => :OP_EQEQ,
		'>>' => :OP_SHIFTRIGHT,
		'||' => :OP_CONCATENATE
	}

	class Lex
		attr_reader :line_counter

		def initialize scanner
			@scanner = scanner
			@line_counter = 1
		end

		def qre pattern  # quote_regexp
			Regexp.new(Regexp.quote(pattern))
		end

		def skip_comment
			unless @scanner.skip(qre('/*')) == 2 then
				raise
			end
			lambda do
				if @scanner.eos? then
					raise "EOF in comment"
				end
				case @scanner.peek 1
				when "\n" then
					@scanner.get_byte
					@line_counter += 1
					redo
				when "\r" then
					@scanner.get_byte
					if @scanner.peek(1) == "\n" then
						@scanner.get_byte
					end
					@line_counter += 1
					redo
				when '/' then
					if @scanner.peek(2) == '/*' then
						skip_comment  # nested comment
					else
						@scanner.get_byte
					end
					redo
				when '*' then
					unless @scanner.peek(2) == '*/' then
						@scanner.get_byte
						redo
					else
						unless @scanner.skip(qre('*/')) == 2 then
							raise
						end
						# break
					end
				else
					@scanner.get_byte
					redo
				end
			end.call
		end

		def proc_const
			unless @scanner.skip(/0/) == 1 then
				raise
			end
			token = '0'
			lambda do
				case @scanner.peek 1
				when 'x', 'X' then
					hex = @scanner.scan /[xX][0-9A-Fa-f]*/
					token << hex
					redo
				when 'o', 'O' then
					oct = @scanner.scan /[oO][0-7]*/
					token << oct
					redo
				when 'b', 'B' then
					bin = @scanner.scan /[bB][01]*/
					token << bin
					redo
				end
			end.call
			[:CONST, token]
		end

		def proc_word token
			if RESERVED.has_key? token then
				raise '"%s": reserved word in line %d' % [token, @line_counter]
			elsif /t_adrs_[0-9][0-9]*/ =~ token then
				raise '"%s": reserved word in line %d' % [token, @line_counter]
			elsif KEYWORDS.has_key? token then
				sym = KEYWORDS[token]
				[sym, token]
			else
				[:NAME, token]
			end
		end

		def next_token
			lambda do
				if @scanner.eos? then
					[false, nil]
				else
					c = @scanner.peek 1
					case c
					when "\t", "\xb", "\f", ' ' then
						@scanner.get_byte
						redo
					when "\n" then
						@scanner.get_byte
						@line_counter += 1
						redo
					when "\r" then
						@scanner.get_byte
						if @scanner.peek(1) == "\n" then
							@scanner.get_byte
						end
						@line_counter += 1
						redo
					when '/' then
						cc = @scanner.peek 2
						if cc == '/*' then
							skip_comment
							redo
						elsif DBOPS.has_key? cc then  # /& /@ /|
							@scanner.skip(qre(cc))
							[DBOPS[cc], cc]
						else
							@scanner.get_byte
							[c, c]
						end
					when '\\' then
						@scanner.get_byte
						[:OP_YEN, c]
					when '#', '&', '(', ')', '+', ',', '.', ';', '@', '[', ']', '^', '{', '}' then
						@scanner.get_byte
						[c, c]
					when ':', '<', '=', '>', '|' then
						cc = @scanner.peek 2
						if DBOPS.has_key? cc then  # := << == >> ||
							@scanner.skip(qre(cc))
							[DBOPS[cc], cc]
						else
							@scanner.get_byte
							[c, c]
						end
					when '0' then
						case @scanner.peek 2
						when '0x', '0X', '0o', '0O', '0b', '0B' then
							proc_const
						else
							@scanner.get_byte
							[:NUM, 0]
						end
					when '1' .. '9' then
						token = @scanner.scan /[1-9][0-9]*/
						[:NUM, token.to_i]
					when 'A' .. 'Z', 'a' .. 'z' then
						token = @scanner.scan /[A-Za-z][0-9A-Z_a-z]*/
						proc_word token
					else
						c = @scanner.get_byte
						raise '"%s": cannot scan in line %d' % [c, @line_counter]
					end
				end
			end.call
		end
	end
end
