#!/usr/local/bin/ruby
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:

require 'yaml'

require 'sfllib'

module SFL
	class Descriptions
		def to_sfl
			l = []
			@descriptions.inject '' do |s, description|
				l << (description.to_sfl 0)
			end
			l.join "\n"
		end
	end

	def SFL.syntax_element_inject l, elems, *opt
		unless elems.empty? then
			l << ( elems.inject '' do |s, e|
				"#{s}#{e.to_sfl *opt}"
			end )
		end
	end

	class ModuleDecl
		def to_sfl indent
			indent_str = ' ' * indent
			l = []
			SFL.syntax_element_inject l, @exterms, indent + 2
			SFL.syntax_element_inject l, @instr_use_args, indent + 2
			"#{indent_str}declare #{@name} {\n#{l.join "\n"}#{indent_str}}\n"
		end
	end

	def SFL.comma_list list
		l = []
		list.each do |entry|
			l << "#{entry.name}"
		end
		l.join ', '
	end

	def SFL.comma_list_term list
		l = []
		list.each do |entry|
			if entry.opt[0] then
				l << "#{entry.name}<#{entry.opt[0]}>"
			else
				l << "#{entry.name}"
			end
		end
		l.join ', '
	end

	def SFL.comma_list_mem list
		l = []
		list.each do |entry|
			if entry.opt[1] then
				l << "#{entry.name}[#{entry.opt[0]}]<#{entry.opt[1]}>"
			else
				l << "#{entry.name}[#{entry.opt[0]}]"
			end
		end
		l.join ', '
	end

	class NameList
		def to_sfl indent
			s = "#{' ' * indent}#{@type.to_s} "
			case @type
			when :input, :output, :bidirect, :bus, :sel, :bus_v, :sel_v, :reg, :reg_ws, :reg_wr
				s << (SFL.comma_list_term @list)
			when :mem
				s << (SFL.comma_list_mem @list)
			else
				s << (SFL.comma_list @list)
			end
			s << ";\n"
			s
		end
	end

	class InstrArg
		def to_sfl indent
			s = "#{' ' * indent}instr_arg #{@name}("
			s << (SFL.comma_list @list)
			s << ");\n"
			s
		end
	end

	class InstrArgEx
		def to_sfl indent
			s = "#{' ' * indent}instr_arg #{@mod_name}.#{@name}("
			s << (SFL.comma_list @list)
			s << ");\n"
			s
		end
	end

	class ModuleDef
		def to_sfl indent
			l = []
			SFL.syntax_element_inject l, @module_decls, 2
			SFL.syntax_element_inject l, @term_elem_defs, 2
			SFL.syntax_element_inject l, @instr_args, 2
			SFL.syntax_element_inject l, @stage_names
			if @common_action then
				l << "  #{@common_action.to_sfl 2}\n"
			end
			SFL.syntax_element_inject l, @instructs
			SFL.syntax_element_inject l, @stages
			s = "module #{@name} {\n#{l.join "\n"}}\n"
		end
	end

	class StageName
		def to_sfl
			s = "  stage_name #{@name} { "
			s = @tasks.inject s do |s, task|
				"#{s}#{task.to_sfl} "
			end
			s = "#{s}}\n"
			s
		end
	end

	class Task
		def to_sfl
			s = "task #{@name}("
			s << (SFL.comma_list @list)
			s << ");"
			s
		end
	end

	class Instruct
		def to_sfl
			"  instruct #{@name} #{@action.to_sfl 2}\n"
		end
	end

	class InstructEx
		def to_sfl
			"  instruct #{@mod_name}.#{@name} #{@action.to_sfl 2}\n"
		end
	end

	class Stage
		def to_sfl
			l = []
			SFL.syntax_element_inject l, @state_names, 4
			SFL.syntax_element_inject l, @segment_names, 4
			if @first_state then
				l << "    first_state #{@first_state};\n"
			end
			if @common_action then
				l << "    #{@common_action.to_sfl 4}\n"
			end
			SFL.syntax_element_inject l, @states, 4
			SFL.syntax_element_inject l, @segments
			"  stage #{@name} {\n#{l.join "\n"}  }\n"
		end
	end

	class Segment
		def to_sfl
			l = []
			SFL.syntax_element_inject l, @state_names, 6
			if @first_state then
				l << "      first_state #{@first_state};\n"
			end
			l << "      #{@common_action.to_sfl 6}\n"
			SFL.syntax_element_inject l, @states, 6
			"    segment #{@name} {\n#{l.join "\n"}    }\n"
		end
	end

	class State
		def to_sfl indent
			indent_str = ' ' * indent
			"#{indent_str}state #{@name} #{@action.to_sfl indent}\n"
		end
	end

	class ActPar
		def to_sfl indent
			indent_str = ' ' * indent
			s = "par {\n"
			s = @actions.inject s do |s, action|
				"#{s}#{indent_str}  #{action.to_sfl indent + 2}\n"
			end
			s = "#{s}#{indent_str}}"
			s
		end
	end

	class ActAlt
		def to_sfl indent
			indent_str = ' ' * indent
			s = "alt {\n"
			s = @gurded_actions.inject s do |s, g_action|
				"#{s}#{indent_str}  #{g_action.to_sfl indent + 2}\n"
			end
			if @else_action then
				s = "#{s}#{indent_str}  else : #{@else_action.to_sfl indent + 2}\n"
			end
			s = "#{s}#{indent_str}}"
			s
		end
	end

	class ActAny
		def to_sfl indent
			indent_str = ' ' * indent
			s = "any {\n"
			s = @gurded_actions.inject s do |s, g_action|
				"#{s}#{indent_str}  #{g_action.to_sfl indent + 2}\n"
			end
			if @else_action then
				s = "#{s}#{indent_str}  else : #{@else_action.to_sfl indent + 2}\n"
			end
			s = "#{s}#{indent_str}}"
			s
		end
	end

	def SFL.exp_to_sfl exp
		if exp.kind_of? String then
			exp
		else
			exp.to_sfl
		end
	end

	class GuardedAct
		def to_sfl indent
			indent_str = ' ' * indent
			"#{SFL.exp_to_sfl @exp} : #{@action.to_sfl indent}"
		end
	end

	class ActIf
		def to_sfl indent
			"if (#{SFL.exp_to_sfl @exp}) #{@action.to_sfl indent}"
		end
	end

	class UActOutput
		def to_sfl indent
			"#{@name} = #{SFL.exp_to_sfl @exp};"
		end
	end

	class UActOutputEx
		def to_sfl indent
			"#{@mod_name}.#{@name} = #{SFL.exp_to_sfl @exp};"
		end
	end

	class UActRegWrite
		def to_sfl indent
			"#{@name} := #{SFL.exp_to_sfl @exp};"
		end
	end

	class UActMemWrite
		def to_sfl indent
			"#{@name}[#{SFL.exp_to_sfl @adr_exp}] := #{SFL.exp_to_sfl @exp};"
		end
	end

	class UActGoto
		def to_sfl indent
			"goto #{@name};"
		end
	end

	class UActCall
		def to_sfl indent
			if @state_name then
				"call #{@seg_name}(#{@state_name});"
			else
				"call #{@seg_name}();"
			end
		end
	end

	class UActReturn
		def to_sfl indent
			"return;"
		end
	end

	class UActGenerate
		def to_sfl indent
			l = []
			@exps.each do |exp|
				l << (SFL.exp_to_sfl exp)
			end
			"generate #{@stage_name}.#{@task_name}(#{l.join ', '});"
		end
	end

	class UActRelay
		def to_sfl indent
			l = []
			@exps.each do |exp|
				l << (SFL.exp_to_sfl exp)
			end
			"relay #{@stage_name}.#{@task_name}(#{l.join ', '});"
		end
	end

	class UActFinish
		def to_sfl indent
			"finish;"
		end
	end

	class UActInstruct
		def to_sfl indent
			l = []
			@exps.each do |exp|
				l << (SFL.exp_to_sfl exp)
			end
			"#{@name}(#{l.join ', '});"
		end
	end

	class UActInstructEx
		def to_sfl indent
			l = []
			@exps.each do |exp|
				l << (SFL.exp_to_sfl exp)
			end
			"#{@mod_name}.#{@name}(#{l.join ', '});"
		end
	end

	class UActNop
		def to_sfl indent
			";"
		end
	end

	class BExp
		def to_sfl
			"#{SFL.exp_to_sfl @unary} #{@op.to_s} #{SFL.exp_to_sfl @exp}"
		end
	end

	class UExp
		def to_sfl
			"#{@op.to_s}#{SFL.exp_to_sfl @unary}"
		end
	end

	class SignExt
		def to_sfl
			"#{@num.to_s}##{SFL.exp_to_sfl @unary}"
		end
	end

	class CutOut
		def to_sfl
			"#{SFL.exp_to_sfl @elem}<#{@num.to_s}>"
		end
	end

	class CutOut2
		def to_sfl
			"#{SFL.exp_to_sfl @elem}<#{@msb_num.to_s}:#{@lsb_num.to_s}>"
		end
	end

	class ElemParen
		def to_sfl
			"(#{SFL.exp_to_sfl @exp})"
		end
	end

	class ElemMemRead
		def to_sfl
			"#{@name}[#{SFL.exp_to_sfl @exp}]"
		end
	end

	class ElemEx
		def to_sfl
			"#{@mod_name}.#{@name}"
		end
	end

	class ElemInstruct
		def to_sfl
			l = []
			@exps.each do |exp|
				l << (SFL.exp_to_sfl exp)
			end
			"#{@instr_name}(#{l.join ', '}).#{@term_name}"
		end
	end

	class ElemInstructEx
		def to_sfl
			l = []
			@exps.each do |exp|
				l << (SFL.exp_to_sfl exp)
			end
			"#{@mod_name}.#{@instr_name}(#{l.join ', '}).#{@term_name}"
		end
	end
end

tree = nil
open ARGV[0] do |file|
	str = file.read
	tree = YAML.load str
end

result = tree.to_sfl
print result
